use dbus::arg;
use std::collections::HashMap;
use std::time::Duration;

#[derive(serde::Serialize)]
pub struct BluetoothInfo {
    pub name: String,
    pub address: String,
    pub alias: String,
    pub connected: bool,
    pub icon: String,
    pub battery: Option<u64>,
}

impl BluetoothInfo {
    fn new(from: HashMap<String, arg::PropMap>) -> Option<Self> {
        let device = from.get("org.bluez.Device1")?;
        let mut result = Self {
            name: device.get("Name")?.0.as_str()?.to_owned(),
            address: device.get("Address")?.0.as_str()?.to_owned(),
            alias: device.get("Alias")?.0.as_str()?.to_owned(),
            connected: matches!(&device.get("Connected")?.0.as_u64(), Some(1)),
            icon: device.get("Icon")?.0.as_str()?.to_owned(),
            battery: None,
        };
        if let Some(battery) = from.get("org.bluez.Battery1") {
            result.battery = Some(battery.get("Percentage")?.0.as_u64()?);
        }
        Some(result)
    }
}

pub fn get_all_bluetooth_devices(
    connection: &dbus::blocking::Connection,
) -> Vec<BluetoothInfo> {
    let proxy = connection.with_proxy("org.bluez", "/", Duration::from_millis(5000));
    let (result,): (HashMap<dbus::Path<'static>, HashMap<String, arg::PropMap>>,) = proxy
        .method_call(
            "org.freedesktop.DBus.ObjectManager",
            "GetManagedObjects",
            (),
        )
        .unwrap_or_default();
    let result: Vec<BluetoothInfo> = result
        .into_iter()
        .filter(|(_, value)| value.contains_key("org.bluez.Device1"))
        .flat_map(|(_, value)| BluetoothInfo::new(value))
        .collect();
    result
}
